/*
 * The tracker.deploy package defines classes for launching and installing Tracker.
 *
 * Copyright (c) 2019  Douglas Brown
 *
 * Tracker is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tracker is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tracker; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA
 * or view the license online at <http://www.gnu.org/copyleft/gpl.html>
 *
 * For additional Tracker information and documentation, please see
 * <http://physlets.org/tracker/>.
 */
package org.opensourcephysics.cabrillo.tracker.deploy;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Map;
import java.nio.charset.Charset;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

import org.opensourcephysics.cabrillo.tracker.Tracker;
import org.opensourcephysics.controls.XML;
import org.opensourcephysics.controls.XMLControl;
import org.opensourcephysics.controls.XMLControlElement;
import org.opensourcephysics.display.OSPRuntime;
import org.opensourcephysics.tools.JREFinder;
import org.opensourcephysics.tools.ResourceLoader;

/**
 * A class to start Tracker. This is the main executable when Tracker is
 * deployed.
 * 
 * @author Douglas Brown
 */
public class TrackerStarter {

	public static final String PREFERRED_TRACKER_JAR = "PREFERRED_TRACKER_JAR"; //$NON-NLS-1$
	public static final String PREFERRED_MEMORY_SIZE = "PREFERRED_MEMORY_SIZE"; //$NON-NLS-1$
//	public static final String PREFERRED_JAVA_VM = "PREFERRED_JAVA_VM"; //$NON-NLS-1$
//	public static final String PREFERRED_VM_BITNESS = "PREFERRED_VM_BITNESS"; //$NON-NLS-1$
//	public static final String PREFERRED_TRACKER_PREFS = "PREFERRED_TRACKER_PREFS"; //$NON-NLS-1$
	public static final String TRACKER_RELAUNCH = "TRACKER_RELAUNCH"; //$NON-NLS-1$	
	public static final String TRACKER_NEW_VERSION = "TRACKER_NEW_VERSION"; //$NON-NLS-1$	
	public static final String LOG_FILE_NAME = "tracker_start.log"; //$NON-NLS-1$
  public static final int DEFAULT_MEMORY_SIZE = 256;
	public static final String PREFS_FILE_NAME = "tracker.prefs"; //$NON-NLS-1$
  
	static String newline = "\n"; //$NON-NLS-1$
	static String encoding = "UTF-8"; //$NON-NLS-1$
	static String exceptions = ""; //$NON-NLS-1$
	static String ffmpegWarning, starterWarning;
	static String trackerHome, userHome, javaHome, ffmpegHome, userDocuments;
	static String startLogPath;
	static FilenameFilter trackerJarFilter = new TrackerJarFilter();
	static File codeBaseDir, starterJarFile;
	static String launchVersionString;
	static String trackerJarPath;
	static int memorySize, preferredMemorySize;
	static String[] executables;
	static String logText = ""; //$NON-NLS-1$
	static String javaCommand = "java"; //$NON-NLS-1$
	static String preferredVM, bundledVM;
	static String snapshot = "-snapshot"; //$NON-NLS-1$
	static boolean debug = false;
	static boolean log = true;
	static boolean relaunching = false;
	static boolean launching = false;
	static int port = 12321;
	static Thread launchThread, exitThread;
	static boolean abortExit;
	static int exitCounter = 0;
	
	static {
		// identify codeBaseDir
		try {
			newline = System.getProperty("line.separator", "\n"); //$NON-NLS-1$ //$NON-NLS-2$
			URL url = TrackerStarter.class.getProtectionDomain().getCodeSource().getLocation();
			java.net.URI uri = url.toURI();
			String path = uri.toString();
			if (path.startsWith("jar:")) { //$NON-NLS-1$
				path = path.substring(4, path.length());
			}
			uri = new java.net.URI(path);
			starterJarFile = new File(uri);
			codeBaseDir = starterJarFile.getParentFile();
			OSPRuntime.setLaunchJarPath(starterJarFile.getAbsolutePath());
		} catch (Exception ex) {
			exceptions += ex.getClass().getSimpleName()
					+ ": " + ex.getMessage() + newline; //$NON-NLS-1$
		}
		// get user home, java home and ffmpeg home
		try {
			userHome = OSPRuntime.getUserHome();
			javaHome = System.getProperty("java.home"); //$NON-NLS-1$
			if (OSPRuntime.isWindows()) {
				userDocuments = new JFileChooser().getFileSystemView().getDefaultDirectory().toString();
			} 
			else {
				userDocuments = userHome + "/Documents"; //$NON-NLS-1$
			}
			if (!new File(userDocuments).exists()) {
				userDocuments = null;
			}
		} catch (Exception ex) {
			exceptions += ex.getClass().getSimpleName()
					+ ": " + ex.getMessage() + newline; //$NON-NLS-1$
		}
	}
	
	/**
	 * Main entry point when used as executable
	 * @param args array of filenames
	 */
	public static void main(final String[] args) {
		relaunching = false;
		logText = ""; //$NON-NLS-1$
		logMessage("launch initiated by user"); //$NON-NLS-1$
		
  	if (OSPRuntime.isMac()) {
  		// create launchThread to instantiate OSXServices and launch Tracker
		  launchThread = new Thread(new Runnable() {
				public void run() {
					// instantiate OSXServices
					String className = "org.opensourcephysics.cabrillo.tracker.deploy.OSXServices"; //$NON-NLS-1$
					try {
						Class<?> OSXClass = Class.forName(className);
						Constructor<?> constructor = OSXClass.getConstructor();
						Object OSXServices = constructor.newInstance();
						Method m = OSXClass.getDeclaredMethod("getStatus", (Class<?>[])null); //$NON-NLS-1$
						Object status = m.invoke(OSXServices, (Object[])null);	
						logMessage(""+status); //$NON-NLS-1$
					} catch (Exception ex) {
						logMessage("OSXServices failed"); //$NON-NLS-1$
					} catch (Error err) {
						logMessage("OSXServices failed"); //$NON-NLS-1$
					}
					// wait a short time for OSXServices to handle openFile event
					// and launch Tracker with file arguments (sets launchThread to null)
					int i = 0;
					while(launchThread!=null && i<5) {
						try {
							Thread.sleep(100);
							i++;
						} catch (InterruptedException e) {
						}
					};
					// launch Tracker with default args if launchThread is not null 
					if (launchThread!=null) {
						launchTracker(args);	
					}
				}
			});
			launchThread.start();				  
		}
  	else {
  		// for Windows and Linux, launch Tracker immediately with default args
			launchTracker(args);					 
  	}
	}

	/**
	 * Launches a new instance of Tracker.
	 * @param args array of filenames. The first arg may be a preferred tracker.jar.
	 */
	public static void launchTracker(String[] args) {
		if (launching) return;
		launching = true;
		logMessage("TrackerStarter running in jre: " + javaHome); //$NON-NLS-1$
		
		launchThread = null;
		
		// look for new version tracker jar as first argument
		if (args!=null && args.length>0 && (args[0].contains("tracker.jar")  //$NON-NLS-1$
					|| (args[0].contains("tracker-") && args[0].contains(".jar")))) { //$NON-NLS-1$ //$NON-NLS-2$
    	System.setProperty(TrackerStarter.PREFERRED_TRACKER_JAR, args[0]);
    	System.setProperty(TrackerStarter.TRACKER_NEW_VERSION, args[0]);
    	String[] newArgs = new String[args.length-1];
    	if (newArgs.length>0) {
    		System.arraycopy(args, 1, newArgs, 0, newArgs.length);
    		args = newArgs;
    	}
    	else args = null;
		}
		
		String argString = null;
		if (args != null && args.length > 0) {
			argString = ""; //$NON-NLS-1$
			for (String next : args) {
				argString += "\"" + next + "\" "; //$NON-NLS-1$ //$NON-NLS-2$
			}
		}
		logMessage("launching with main arguments: " + argString); //$NON-NLS-1$
		

		// find Tracker home
		try {
			trackerHome = findTrackerHome(true);
		} catch (Exception ex) {
			exceptions += ex.getClass().getSimpleName()
					+ ": " + ex.getMessage() + newline; //$NON-NLS-1$
		}
		if (trackerHome == null) {
			exitGracefully(null);
		}

		// find FFMPeg home
		try {
			ffmpegHome = findFFMPegHome(trackerHome, true);
		} catch (Exception ex) {
			exceptions += ex.getClass().getSimpleName()
					+ ": " + ex.getMessage() + newline; //$NON-NLS-1$
		}

		// load preferences
		loadPreferences();

		// determine which tracker jar to launch
		String jarPath = null;		
		try {
			jarPath = getTrackerJarPath();
		} catch (Exception ex) {
			exceptions += ex.getClass().getSimpleName()
					+ ": " + ex.getMessage() + newline; //$NON-NLS-1$
		}
		if (jarPath == null) {
			exitGracefully(null);
		}

		// launch Tracker
		boolean launched = true;
		try {
			memorySize = preferredMemorySize;
			startTracker(jarPath, args);
		} catch (Exception ex) {
			launched = false;
			exceptions += ex.getClass().getSimpleName()
					+ ": " + ex.getMessage() + newline; //$NON-NLS-1$
		} catch (Error er) {
			launched = false;
			exceptions += er.getClass().getSimpleName()
					+ ": " + er.getMessage() + newline; //$NON-NLS-1$
		}

		if (!launched) {
			exitGracefully(jarPath);
		}
	}


	
	/**
	 * Relaunches a new instance of Tracker.
	 * @param args array of filenames
	 * @param secondTry true to flag this as a second try in the new process
	 */
	public static void relaunch(final String[] args, boolean secondTry) {
		relaunching = secondTry;
		launching = false;
		Runnable runner = new Runnable() {
			public void run() {
				logText = ""; //$NON-NLS-1$
				logMessage("relaunch initiated by Tracker"); //$NON-NLS-1$
				launchTracker(args);
			}
		};
		new Thread(runner).start();
	}

	/**
	 * Finds the Tracker home directory and sets/returns the static variable trackerHome.
	 * 
	 * @param writeToLog true to write the results to the start log
	 */
	public static String findTrackerHome(boolean writeToLog) throws Exception {
		if (trackerHome!=null) return trackerHome;
		// first determine if code base directory is trackerHome
		if (codeBaseDir != null) {
			if (writeToLog) {
				logMessage("TrackerStarter jar: " + starterJarFile); //$NON-NLS-1$
			}
			// accept if the directory has any tracker jars in it
			try {
				String[] fileNames = codeBaseDir.list(trackerJarFilter);
				if (fileNames != null && fileNames.length > 0) {
					trackerHome = codeBaseDir.getPath();
					if (writeToLog) logMessage("code base accepted as trackerhome based on contents"); //$NON-NLS-1$
				}
			} catch (Exception ex) {
				exceptions += ex.getClass().getSimpleName()
						+ ": " + ex.getMessage() + newline; //$NON-NLS-1$
			}
		}

		// if needed, try another way to see if current directory has tracker.jar
		if (trackerHome == null) {
			File file = new File((String) null, "tracker.jar"); //$NON-NLS-1$
			String dir = file.getAbsoluteFile().getParent();
			if (fileExists(file.getAbsolutePath())) {
				trackerHome = dir;
				if (writeToLog) logMessage("parent directory accepted as trackerhome based on contents"); //$NON-NLS-1$
			}
		}

		// if not found locally, look for (legacy) environment variable
		if (trackerHome == null) {
			try {
				trackerHome = System.getenv("TRACKER_HOME"); //$NON-NLS-1$
			} catch (Exception ex) {
				exceptions += ex.getClass().getSimpleName()
						+ ": " + ex.getMessage() + newline; //$NON-NLS-1$
			}
			if (writeToLog) logMessage("environment variable TRACKER_HOME: " + trackerHome); //$NON-NLS-1$
			if (trackerHome != null && !fileExists(trackerHome)) {
				trackerHome = null;
				if (writeToLog) logMessage("TRACKER_HOME directory no longer exists"); //$NON-NLS-1$
			}
		}

		if (trackerHome == null)
			throw new NullPointerException("trackerhome not found"); //$NON-NLS-1$
		if (writeToLog) logMessage("using trackerhome: " + trackerHome); //$NON-NLS-1$
		
		return trackerHome;
	}
	
	/**
	 * Finds and loads the preferences file into an XMLControl.
	 * 
	 * @return the loaded XMLControl, or null if no preferences file found
	 */
	public static XMLControl findPreferences() {
  	// look for all prefs files in OSPRuntime.getDefaultSearchPaths()
		// and in current directory
    Map<File, XMLControl> controls = new HashMap<File, XMLControl>();
  	File firstFileFound = null, newestFileFound = null;
  	long modified = 0;
  	
  	for (int i=0; i<2; i++) {
			String prefsFileName = PREFS_FILE_NAME;
			if (i==1) {
				// add leading dot to fileName
				prefsFileName = "."+prefsFileName; //$NON-NLS-1$
			}
  		for (String path: OSPRuntime.getDefaultSearchPaths()) {
	      String prefsPath = new File(path, prefsFileName).getAbsolutePath();
	      XMLControl control = new XMLControlElement(prefsPath);
	      if (!control.failedToRead() && control.getObjectClassName().endsWith("Preferences")) { //$NON-NLS-1$
	      	File file = new File(prefsPath);
	      	if (file.lastModified()>modified+50) {
	      		newestFileFound = file;
		      	modified = file.lastModified();
	      	}
	      	controls.put(file, control);
	      	if (firstFileFound==null) {
		      	firstFileFound = file;
	      	}
	      }
  		}
  		// look in current directory
      String prefsPath = new File(prefsFileName).getAbsolutePath();
      XMLControl control = new XMLControlElement(prefsPath);
      if (!control.failedToRead() && control.getObjectClassName().endsWith("Preferences")) { //$NON-NLS-1$
      	File file = new File(prefsPath);
      	if (file.lastModified()>modified+50) {
      		newestFileFound = file;
	      	modified = file.lastModified();
      	}
      	controls.put(file, control);
      	if (firstFileFound==null) {
	      	firstFileFound = file;
      	}
      }
  	}
  	// replace first file with newest if different
  	if (newestFileFound!=firstFileFound) {
  		ResourceLoader.copyAllFiles(newestFileFound, firstFileFound);
  		controls.put(firstFileFound, controls.get(newestFileFound));
  	}
		
  	// return control associated with first file found
  	if (firstFileFound!=null) {
  		XMLControl control = controls.get(firstFileFound);
  		control.setValue("prefsPath", firstFileFound.getAbsolutePath()); //$NON-NLS-1$
    	return control;
  	}
  	return null;
	}

	/**
	 * Finds the FFMPeg home directory and sets/returns the static variable ffmpegHome.
	 * 
	 * @param trackerHome the Tracker home directory (may be null)
	 * @param writeToLog true to write results to the start log
	 */
	public static String findFFMPegHome(String trackerHome, boolean writeToLog) throws Exception {
		// first see if ffmpegHome is child or sibling of trackerHome
		if (trackerHome!=null) {
			File trackerHomeDir = new File(trackerHome);
			File f = new File(trackerHomeDir, "FFMPeg"); //$NON-NLS-1$
			if (!f.exists() || !f.isDirectory()) {
				f = new File(trackerHomeDir.getParentFile(), "FFMPeg"); //$NON-NLS-1$
			}
			if ((!f.exists()||!f.isDirectory()) && OSPRuntime.isMac()) {
				f = new File("/usr/local/ffmpeg"); //$NON-NLS-1$
			}
			if (f.exists() && f.isDirectory()) {
				ffmpegHome = f.getPath();
				if (writeToLog) logMessage("ffmpegHome found relative to trackerhome: "+ffmpegHome); //$NON-NLS-1$
			}
		}
		
		// if not found, check OSP preferences
		if (ffmpegHome==null) {
			ffmpegHome = (String)OSPRuntime.getPreference("FFMPEG_HOME"); //$NON-NLS-1$
			if (writeToLog) logMessage("osp.prefs FFMPEG_HOME: " + ffmpegHome); //$NON-NLS-1$
			if (ffmpegHome!=null && !fileExists(ffmpegHome)) {
				ffmpegHome = null;
				if (writeToLog) logMessage("FFMPEG_HOME directory no longer exists"); //$NON-NLS-1$
			}	
		}

		// if not yet found, look for ffmpegHome in environment variable
		if (ffmpegHome==null) {
			try {
				ffmpegHome = System.getenv("FFMPEG_HOME"); //$NON-NLS-1$
			} catch (Exception ex) {
				exceptions += ex.getClass().getSimpleName()
						+ ": " + ex.getMessage() + newline; //$NON-NLS-1$
			}
			if (writeToLog) logMessage("environment variable FFMPEG_HOME: " + ffmpegHome); //$NON-NLS-1$
			if (ffmpegHome!=null && !fileExists(ffmpegHome)) {
				ffmpegHome = null;
				if (writeToLog) logMessage("FFMPEG_HOME directory no longer exists"); //$NON-NLS-1$
			}			
		}

		if (ffmpegHome==null)
			throw new NullPointerException("ffmpegHome not found"); //$NON-NLS-1$
		if (writeToLog) logMessage("using ffmpegHome: " + ffmpegHome); //$NON-NLS-1$
		return ffmpegHome;
	}
	
	/**
	 * Finds the bundled Java vm, if any.
	 */
	public static String findBundledVM() {
		if (bundledVM!=null) return bundledVM;
		try {
			findTrackerHome(false);
		} catch (Exception e) {}
		
		File jre = null;
		if (OSPRuntime.isWindows()) {
			jre = JREFinder.getFinder().getDefaultJRE(32, trackerHome, false);
		}
		else if (OSPRuntime.isMac()) {
			File home = new File(trackerHome);
			String path = home.getParent()+"/PlugIns/Java.runtime"; //$NON-NLS-1$
			jre = JREFinder.getFinder().getDefaultJRE(64, path, false);
		}
		else {
			jre = JREFinder.getFinder().getDefaultJRE(OSPRuntime.getVMBitness(), trackerHome, false);
		}
		if (jre != null && jre.exists()) {
			return jre.getPath();
		}
		return null;
	}


	/**
	 * Exits gracefully by giving information to the user.
	 */
	private static void exitGracefully(String jarPath) {
		if (exitThread!=null) abortExit = true;
		if (exceptions.equals("")) //$NON-NLS-1$
			exceptions = "None"; //$NON-NLS-1$
		String startLogLine = ""; //$NON-NLS-1$
		if (startLogPath!=null) {
			startLogLine = "For more information see " + startLogPath + newline; //$NON-NLS-1$
		}
		if (jarPath != null) {
			JOptionPane
					.showMessageDialog(
							null,
							"Tracker could not be started due to the problem(s) listed below." + newline //$NON-NLS-1$
									+ "However, you may be able to start it by double-clicking the file" + newline //$NON-NLS-1$
									+ jarPath
									+ "." + newline + newline //$NON-NLS-1$
									+ startLogLine
									+ "For trouble-shooting or to download the latest installer," + newline //$NON-NLS-1$
									+ "please see http://physlets.org/tracker/." + newline + newline //$NON-NLS-1$
									+ "Problems:" + newline + exceptions, //$NON-NLS-1$
							"TrackerStarter Vers " + Tracker.VERSION + ": Error Starting Tracker", //$NON-NLS-1$ //$NON-NLS-2$
							JOptionPane.ERROR_MESSAGE);
		} else {
			if (trackerHome == null) {
				if (codeBaseDir != null) {
					JOptionPane
							.showMessageDialog(
									null,
									"It appears you have an incomplete Tracker installation, since" + newline //$NON-NLS-1$
											+ "no directory named \"Tracker\" could be found and " + newline //$NON-NLS-1$
											+ "no tracker.jar or tracker-x.xx.jar file exists in " + newline //$NON-NLS-1$
											+ codeBaseDir
											+ newline
											+ newline
											+ startLogLine
											+ "For trouble-shooting or to download the latest installer," + newline //$NON-NLS-1$
											+ "please see http://physlets.org/tracker/." + newline + newline //$NON-NLS-1$
											+ "Problems:" + newline + exceptions, //$NON-NLS-1$
									"TrackerStarter Vers " + Tracker.VERSION + ": Error Starting Tracker", //$NON-NLS-1$ //$NON-NLS-2$
									JOptionPane.ERROR_MESSAGE);
				} else {
					JOptionPane
							.showMessageDialog(
									null,
									"It appears you have an incomplete Tracker installation, since" + newline //$NON-NLS-1$
											+ "no directory named \"Tracker\" could be found and " + newline //$NON-NLS-1$
											+ "no tracker.jar or tracker-x.xx.jar file exists in the current directory." + newline + newline //$NON-NLS-1$
											+ startLogLine
											+ "For trouble-shooting or to download the latest installer," + newline //$NON-NLS-1$
											+ "please see http://physlets.org/tracker/." + newline + newline //$NON-NLS-1$
											+ "Problems:" + newline + exceptions, //$NON-NLS-1$
									"TrackerStarter Vers " + Tracker.VERSION + ": Error Starting Tracker", //$NON-NLS-1$ //$NON-NLS-2$
									JOptionPane.ERROR_MESSAGE);
				}
			} else {
				String jarHome = OSPRuntime.isMac() ? codeBaseDir.getAbsolutePath()
						: trackerHome;
				JOptionPane
						.showMessageDialog(
								null,
								"No tracker.jar or tracker-x.xx.jar was found in" + newline //$NON-NLS-1$
										+ jarHome
										+ newline
										+ newline
										+ startLogLine
										+ "For trouble-shooting or to download the latest installer," + newline //$NON-NLS-1$
										+ "please see http://physlets.org/tracker/." + newline + newline //$NON-NLS-1$
										+ "Problems:" + newline + exceptions, //$NON-NLS-1$
								"TrackerStarter Vers " + Tracker.VERSION + ": Error Starting Tracker", //$NON-NLS-1$ //$NON-NLS-2$
								JOptionPane.ERROR_MESSAGE);
			}

		}
		writeUserLog();
		writeCodeBaseLog();
		System.exit(0);
	}

	/**
	 * Loads preferences from a preferences file.
	 */
	private static void loadPreferences() {
		trackerJarPath = null;
		boolean loaded = false;
		
		XMLControl prefsXMLControl = findPreferences();
		if (prefsXMLControl==null) {
			logMessage("no preferences file found"); //$NON-NLS-1$    
			return;
		}
		String prefsPath = prefsXMLControl.getString("prefsPath"); //$NON-NLS-1$
		
		// now read the preferences from the prefsXMLControl
		// but also check environment preferences which trump prefs file
		if (!prefsXMLControl.failedToRead()) {
			logMessage("loading starter preferences from: " + prefsPath); //$NON-NLS-1$
			
			String jar = null; // preferred jar name to be determined
			
			// preferred tracker jar
			String systemProperty = System.getProperty(PREFERRED_TRACKER_JAR);

			if (jar==null && systemProperty!=null) {
				loaded = true;
				trackerJarPath = systemProperty;
				jar = XML.getName(trackerJarPath);
				logMessage("system property "+PREFERRED_TRACKER_JAR+" = " + systemProperty); //$NON-NLS-1$ //$NON-NLS-2$
			}
			else if (jar==null && prefsXMLControl.getPropertyNames().contains("tracker_jar")) { //$NON-NLS-1$
				loaded = true;
				jar = prefsXMLControl.getString("tracker_jar"); //$NON-NLS-1$
			}

			if (jar!=null && !jar.equals("tracker.jar")) { //$NON-NLS-1$
				int dot = jar.indexOf(".jar"); //$NON-NLS-1$
				String ver = jar.substring(8, dot);
				String versionStr = ver;
    		int n = ver.toLowerCase().indexOf(snapshot);
    		if (n>-1) {
    			ver = ver.substring(0, n);
    		}
				if (new Tracker.Version(ver).isValid()) {
					launchVersionString = versionStr;
					logMessage("preferred version: " + launchVersionString); //$NON-NLS-1$
				} 
				else {
					logMessage("version number not valid: " + ver); //$NON-NLS-1$
				}
			}


			// preferred java vm
			preferredVM = null;
			if (prefsXMLControl.getPropertyNames().contains("java_vm")) { //$NON-NLS-1$
				loaded = true;
				preferredVM = prefsXMLControl.getString("java_vm"); //$NON-NLS-1$
			}
			if (preferredVM!=null) {
				logMessage("preferred java VM: " + preferredVM); //$NON-NLS-1$
				File javaFile = OSPRuntime.getJavaFile(preferredVM);
				if (javaFile != null) {
					javaCommand = XML.stripExtension(javaFile.getPath());
				} 
				else {
					logMessage("preferred java VM invalid"); //$NON-NLS-1$
					preferredVM = null;
				}
			}
			if (preferredVM==null) {
					// look for bundled jre
				bundledVM = findBundledVM();
				if (bundledVM != null) {
					File javaFile = OSPRuntime.getJavaFile(bundledVM);
					if (javaFile!=null) {
						preferredVM = bundledVM;
						logMessage("no preferred java VM, using bundled VM: "+bundledVM); //$NON-NLS-1$
						javaCommand = XML.stripExtension(javaFile.getPath());
					}
				} 
				else {
					logMessage("no preferred or bundled java VM, using default"); //$NON-NLS-1$
				}
			}


			// preferred executables to run prior to starting Tracker
			if (prefsXMLControl.getPropertyNames().contains("run")) { //$NON-NLS-1$
				loaded = true;
				executables = (String[]) prefsXMLControl.getObject("run"); //$NON-NLS-1$
				for (int i = 0; i < executables.length; i++) {
					String app = executables[i];
					if (app == null)
						continue;
					File runFile = new File(trackerHome, app); // app is relative address
					if (!runFile.exists()) {
						runFile = new File(app); // app is absolute address
						if (!runFile.exists()) {
							runFile = null;
							logMessage("executable file not found: " + app); //$NON-NLS-1$
						}
					}
					if (runFile != null)
						try {
							logMessage("executing " + runFile.getAbsolutePath()); //$NON-NLS-1$
							ProcessBuilder pb = new ProcessBuilder(runFile.getAbsolutePath());
							pb.directory(new File(trackerHome));
							Process p = pb.start();
							p.waitFor();
						} catch (Exception ex) {
							logMessage("execution failed: " + ex.getClass().getSimpleName() + " " + ex.getMessage()); //$NON-NLS-1$ //$NON-NLS-2$
						}
				}
			}
			
			// preferred memory size
			preferredMemorySize = 0; // default
			systemProperty = System.getProperty(PREFERRED_MEMORY_SIZE);
			if (systemProperty!=null) {
				loaded = true;
				try {
					preferredMemorySize = Integer.parseInt(systemProperty);
					logMessage("system property "+PREFERRED_MEMORY_SIZE+" = " + systemProperty); //$NON-NLS-1$ //$NON-NLS-2$
				} catch (NumberFormatException e) {
				}
			}
			else if (prefsXMLControl.getPropertyNames().contains("memory_size")) { //$NON-NLS-1$
				preferredMemorySize = prefsXMLControl.getInt("memory_size"); //$NON-NLS-1$
			}
			if (preferredMemorySize>0) {
				logMessage("preferred memory size: " + preferredMemorySize + " MB"); //$NON-NLS-1$ //$NON-NLS-2$				
			}
			else {
				preferredMemorySize = DEFAULT_MEMORY_SIZE;
				logMessage("using default memory size: " + preferredMemorySize + " MB"); //$NON-NLS-1$ //$NON-NLS-2$				
			}
			

			if (!loaded)
				logMessage("no starter preferences found in " + prefsPath); //$NON-NLS-1$      		
		}
	}

	/**
	 * Gets the preferred tracker jar path.
	 * 
	 * @return the path, or null if none found
	 */
	private static String getTrackerJarPath() throws Exception {
		if (trackerJarPath!=null) {
			return trackerJarPath;
		}
		String jarPath = null;
		String jarHome = OSPRuntime.isMac() ? codeBaseDir.getAbsolutePath()
				: trackerHome;
		if (OSPRuntime.isMac()) {
			logMessage("Mac OSX: looking for tracker jars in " + jarHome); //$NON-NLS-1$
		} else if (OSPRuntime.isWindows()) {
			logMessage("Windows: looking for tracker jars in " + jarHome); //$NON-NLS-1$
		} else {
			logMessage("Linux: looking for tracker jars in " + jarHome); //$NON-NLS-1$
		}
		try {
			File dir = new File(jarHome);
			String[] fileNames = dir.list(trackerJarFilter);
			if (fileNames != null && fileNames.length > 0) {
				String s = "tracker jars found: "; //$NON-NLS-1$
				for (String next : fileNames) {
					s += next + ", "; //$NON-NLS-1$
				}
				logMessage(s.substring(0, s.length() - 2));
				String defaultJar = null;
				String numberedJar = null;
				Tracker.Version newestVersion = null;
				for (int i = 0; i < fileNames.length; i++) {
					if ("tracker.jar".equals(fileNames[i].toLowerCase())) {//$NON-NLS-1$
						defaultJar = fileNames[i];
					}
					try {
						String vers = fileNames[i].substring(8);
						vers = vers.substring(0, vers.length() - 4);
						String versionStr = vers;
		    		int n = vers.toLowerCase().indexOf(snapshot);
		    		if (n>-1) {
		    			vers = vers.substring(0, n);
		    		}

		    		Tracker.Version v = new Tracker.Version(vers);
						if (v.isValid()) {
							if (versionStr.equals(launchVersionString)) {
								File file = new File(jarHome, fileNames[i]);
								logMessage("using tracker jar: " + file.getAbsolutePath()); //$NON-NLS-1$
								return file.getAbsolutePath();
							}
							if (newestVersion==null || newestVersion.compareTo(v)<0) {
								newestVersion = v;
								numberedJar = fileNames[i];
							}
						}
					} catch (Exception ex) {
					}
				}
				jarPath = defaultJar != null ? defaultJar : numberedJar;
			}
		} catch (Exception ex) { // if file access fails, try unnumbered tracker.jar
			exceptions += ex.getClass().getSimpleName()
					+ ": " + ex.getMessage() + newline; //$NON-NLS-1$
			logMessage(ex.toString());
			jarPath = "tracker.jar"; //$NON-NLS-1$
		}
		if (jarPath != null) {
			// look in jarHome
			File file = new File(jarHome, jarPath);
			if (file.exists()) {
				logMessage("using tracker jar: " + file.getAbsolutePath()); //$NON-NLS-1$
				return file.getAbsolutePath();
			}
		}
		throw new NullPointerException("No Tracker jar files found in " + jarHome); //$NON-NLS-1$
	}

	/**
	 * Launches the specified tracker jar with a list of arguments
	 * 
	 * @param jarPath the path to the tracker jar
	 * @param args the arguments (may be null)
	 */
	private static void startTracker(String jarPath, String[] args)
			throws Exception {

		String newVersionURL = System.getProperty(TRACKER_NEW_VERSION);
		// pig turn this off
//		// if new version has been installed, run in bundled JRE
//		if (newVersionURL!=null &&  bundledVM != null) {
//			File javaFile = OSPRuntime.getJavaFile(bundledVM);
//			if (javaFile!=null) {
//				javaCommand = XML.stripExtension(javaFile.getPath());
//			} 			
//		}
		// assemble the command
		final ArrayList<String> cmd = new ArrayList<String>();
		if (javaCommand.equals("java") && javaHome != null) { //$NON-NLS-1$
			javaCommand = XML.forwardSlash(javaHome) + "/bin/java"; //$NON-NLS-1$
		}
		cmd.add(javaCommand);

		if (memorySize > 0) {
			cmd.add("-Xms32m"); //$NON-NLS-1$
			cmd.add("-Xmx" + memorySize + "m"); //$NON-NLS-1$ //$NON-NLS-2$
		}
		if (OSPRuntime.isMac()) {
			cmd.add("-Xdock:name=Tracker"); //$NON-NLS-1$
		}
		
		cmd.add("-jar"); //$NON-NLS-1$
		cmd.add(jarPath);
		if (args != null && args.length > 0)
			for (String next : args) {
				if (next != null)
					cmd.add(next);
			}

		// prepare to execute the command
		ProcessBuilder builder = new ProcessBuilder(cmd);
		
		// set environment variables for new process
		Map<String, String> env = builder.environment();
//		String portVar = "TRACKER_PORT"; //$NON-NLS-1$
//		env.put(portVar, String.valueOf(port));
//		if (logText.indexOf(portVar)==-1) {
//			showDebugMessage("setting environment variable "+portVar+" = " + String.valueOf(port)); //$NON-NLS-1$ //$NON-NLS-2$
//		}
		
		if (memorySize<preferredMemorySize) {
			env.put("MEMORY_SIZE", String.valueOf(memorySize)); //$NON-NLS-1$
			logMessage("setting environment variable MEMORY_SIZE = " + String.valueOf(memorySize)); //$NON-NLS-1$ 
		}
		else env.remove("MEMORY_SIZE"); //$NON-NLS-1$ 

		// remove ffmpeg warnings that may have been set by previous versions of TrackerStarter
		env.remove("FFMPEG_WARNING"); //$NON-NLS-1$ 
		if (starterWarning!=null) {
			env.put("STARTER_WARNING", starterWarning); //$NON-NLS-1$ 
		}
		else env.remove("STARTER_WARNING"); //$NON-NLS-1$ 
		
		// add TRACKER_HOME, FFMPEG_HOME and PATH to environment
		if (trackerHome!=null) { 
			env.put("TRACKER_HOME", trackerHome); //$NON-NLS-1$ 
			logMessage("setting TRACKER_HOME = " + trackerHome); //$NON-NLS-1$
		}
		if (ffmpegHome!=null && new File(ffmpegHome).exists()) {
			env.put("FFMPEG_HOME", ffmpegHome); //$NON-NLS-1$ 
			logMessage("setting FFMPEG_HOME = " + ffmpegHome); //$NON-NLS-1$

			String pathEnvironment = OSPRuntime.isWindows()? "Path":  //$NON-NLS-1$
				OSPRuntime.isMac()? "DYLD_LIBRARY_PATH": "LD_LIBRARY_PATH"; //$NON-NLS-1$ //$NON-NLS-2$
			
			String subdir = OSPRuntime.isWindows()? "bin": "lib"; //$NON-NLS-1$ //$NON-NLS-2$
			String ffmpegPath = ffmpegHome+File.separator+subdir;
			if (new File(ffmpegPath).exists()) {
				// get current PATH
				String pathValue = env.get(pathEnvironment);
				if (pathValue==null) pathValue = ""; //$NON-NLS-1$
				
				// add ffmpeg path at beginning of current PATH
				if (!pathValue.startsWith(ffmpegPath)) {
					pathValue = ffmpegPath+File.pathSeparator+pathValue;
				}
				
				env.put(pathEnvironment, pathValue);
				logMessage("setting "+pathEnvironment+" = " + pathValue); //$NON-NLS-1$ //$NON-NLS-2$
			}			
		}
		
		if (ffmpegHome!=null && new File(ffmpegHome).exists()) {
			env.put("FFMPEG_HOME", ffmpegHome); //$NON-NLS-1$ 
			logMessage("setting FFMPEG_HOME = " + ffmpegHome); //$NON-NLS-1$
		}
		
		// add TRACKER_RELAUNCH to process environment if relaunching
		if (relaunching) {
			env.put(TRACKER_RELAUNCH, "true"); //$NON-NLS-1$
		}
		else env.remove(TRACKER_RELAUNCH);
		
		// add TRACKER_NEW_VERSION to process environment if launching a new version
		if (newVersionURL!=null) {
			logMessage("setting "+TRACKER_NEW_VERSION+" = " + newVersionURL); //$NON-NLS-1$ //$NON-NLS-2$ 
			env.put(TRACKER_NEW_VERSION, newVersionURL);
		}
		else env.remove(TRACKER_NEW_VERSION);
		
		// assemble command message for log
		String message = ""; //$NON-NLS-1$
		for (String next: cmd) {
			message += next + " "; //$NON-NLS-1$
		}
		logMessage("executing command: " + message); //$NON-NLS-1$ 

		// write the user tracker_start log and set environment variable
		startLogPath = writeUserLog();
		if (startLogPath!=null)
			env.put("START_LOG", startLogPath); //$NON-NLS-1$
		
		// write codeBase tracker_start log
		writeCodeBaseLog();

		// start exit thread that waits a second before exiting 
		// to give time to start the new process
		exitCounter = 0;
		if (exitThread==null) {
			exitThread = new Thread(new Runnable() {
				public void run() {
					abortExit = false;
					while (exitCounter<10) {
						try {
							if (abortExit) return;
							Thread.sleep(100);
							exitCounter++;
						} catch (InterruptedException e) {
						}
					}					
					System.exit(0);
				}
			});
			exitThread.setDaemon(true);
			exitThread.start();
		}
		
		// start the new Tracker process and wait for it to finish
		// note that successful process should not return until Tracker is exited
		final Process process = builder.start();
		int result = process.waitFor();
		
		// if process returns immediately with exit code 1, log it's error and input streams
		if (result > 0) {
      String errors = ""; //$NON-NLS-1$
			try {
	    	BufferedReader reader = new BufferedReader(new InputStreamReader(process.getErrorStream()));
	      String s = reader.readLine();
	    	while (s != null && s.length()>0) {
	    		errors += "\n      "+s; //$NON-NLS-1$
		      s = reader.readLine();
	    	}
	    	reader.close();
	    	reader = new BufferedReader(new InputStreamReader(process.getInputStream()));
	      s = reader.readLine();
	    	while (s != null && s.length()>0) {
	    		errors += "\n      "+s; //$NON-NLS-1$
		      s = reader.readLine();
	    	}
	    	reader.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
			
	    logMessage("failed to start with memory size "+memorySize+"MB due to the following errors:"+errors); //$NON-NLS-1$ //$NON-NLS-2$
    	
	    // if process failed due to excessive memory size, reduce size and try again
	    if (errors.indexOf("heap")>-1) { //$NON-NLS-1$
    		memorySize *= 0.95;
    		if (memorySize<64) {
    			exceptions += errors + newline;
    			exitGracefully(jarPath);
    		}
    		logMessage("try to start with smaller memory size "+memorySize+"MB"); //$NON-NLS-1$ //$NON-NLS-2$
    		startTracker(jarPath, args);
    	}
	    else {
  			exceptions += errors + newline;
  			exitGracefully(jarPath);
	    }
		}
		else {
			// should never get here--exits via timer
			System.exit(0);
		}
	}

	private static String writeUserLog() {
		if ("".equals(logText) || trackerHome==null) { //$NON-NLS-1$
			return null;
		}	

		// define "user log" file
		File file = null;
		// check if can write to OSP.prefs directory
		File ospPrefsFile = OSPRuntime.getPreferencesFile();
		if (ospPrefsFile != null && ospPrefsFile.getParentFile().canWrite()) {
			file = new File(ospPrefsFile.getParentFile(), LOG_FILE_NAME);
		}
		// check if can write to Tracker file in user documents
		if (file == null && userDocuments != null && new File(userDocuments+"/Tracker").canWrite()) { //$NON-NLS-1$
			file = new File(userDocuments+"/Tracker", LOG_FILE_NAME); //$NON-NLS-1$
		}		
		// check if can write to tracker home
		if (file==null && new File(trackerHome).canWrite()) {
			file = new File(trackerHome, LOG_FILE_NAME);
		}
		
		if (file==null) return null;
		
		addLogHeader();
		logMessage("writing user start log to "+file.getAbsolutePath()); //$NON-NLS-1$

		try {
			FileOutputStream stream = new FileOutputStream(file);
			Charset charset = Charset.forName(encoding);
			OutputStreamWriter out = new OutputStreamWriter(stream, charset);
			BufferedWriter writer = new BufferedWriter(out);
			writer.write(logText);
			writer.flush();
			writer.close();
		} catch (IOException ex) {
			return null;
		}
		return file.getAbsolutePath();
	}
	
	private static void addLogHeader() {
		if (!logText.startsWith("TrackerStarter")) { //$NON-NLS-1$
			SimpleDateFormat sdf = new SimpleDateFormat("HH:mm:ss  MMM dd yyyy"); //$NON-NLS-1$
			Calendar cal = Calendar.getInstance();
			logText = "TrackerStarter version " + Tracker.VERSION + "  " //$NON-NLS-1$ //$NON-NLS-2$
					+ sdf.format(cal.getTime()) + newline + newline + logText;
		}		
	}

	private static void writeCodeBaseLog() {
		// writes log file to codeBaseDir 
		if (codeBaseDir!=null && codeBaseDir.canWrite()) {
			addLogHeader();
			File file = new File(codeBaseDir, LOG_FILE_NAME);
			logMessage("writing code base start log "+file.getAbsolutePath()); //$NON-NLS-1$
			try {
				FileOutputStream stream = new FileOutputStream(file);
				Charset charset = Charset.forName(encoding);
				OutputStreamWriter out = new OutputStreamWriter(stream, charset);
				BufferedWriter writer = new BufferedWriter(out);
				writer.write(logText);
				writer.flush();
				writer.close();
			} catch (IOException ex) {
			}
		}
	}

	private static boolean fileExists(String path) {
		File file = new File(path);
		try {
			if (file.exists()) {
				return true;
			}
		} catch (Exception ex) {
			exceptions += ex.getClass().getSimpleName()
					+ ": " + ex.getMessage() + newline; //$NON-NLS-1$
		} catch (Error er) {
			exceptions += er.getClass().getSimpleName()
					+ ": " + er.getMessage() + newline; //$NON-NLS-1$
		}
		return false;
	}

	private static void logMessage(String message) {
		if (log) {
			logText += " - " + message + newline; //$NON-NLS-1$
		}
		if (debug) {
			System.out.println(message);
		}
	}

}
